import os
import string
import re
import random
import math
import pathlib

import torch
import torch.nn as nn
from torch.autograd import Variable
from torch import optim
import torch.nn.functional as F
from torchtext import data
from torchtext import datasets

def tokenizer(text):
    return list(text.lower())

class Encoder(nn.Module):
    def __init__(self,
                 input_dim,
                 hid_dim,
                 n_layers,
                 n_heads,
                 pf_dim,
                 dropout,
                 device,
                 max_length = 100):
        super().__init__()

        self.device = device

        self.tok_embedding = nn.Embedding(input_dim, hid_dim)
        self.pos_embedding = nn.Embedding(max_length, hid_dim)

        self.layers = nn.ModuleList([EncoderLayer(hid_dim,
                                                  n_heads,
                                                  pf_dim,
                                                  dropout,
                                                  device)
                                     for _ in range(n_layers)])

        self.dropout = nn.Dropout(dropout)

        self.scale = torch.sqrt(torch.FloatTensor([hid_dim])).to(device)

    def forward(self, src, src_mask):

        #src = [batch size, src len]
        #src_mask = [batch size, src len]

        batch_size = src.shape[0]
        src_len = src.shape[1]

        pos = torch.arange(0, src_len).unsqueeze(0).repeat(batch_size, 1).to(self.device)

        #pos = [batch size, src len]

        src = self.dropout((self.tok_embedding(src) * self.scale) + self.pos_embedding(pos))

        #src = [batch size, src len, hid dim]

        for layer in self.layers:
            src = layer(src, src_mask)

        #src = [batch size, src len, hid dim]

        return src

class EncoderLayer(nn.Module):
    def __init__(self,
                 hid_dim,
                 n_heads,
                 pf_dim,
                 dropout,
                 device):
        super().__init__()

        self.self_attn_layer_norm = nn.LayerNorm(hid_dim)
        self.ff_layer_norm = nn.LayerNorm(hid_dim)
        self.self_attention = MultiHeadAttentionLayer(hid_dim, n_heads, dropout, device)
        self.positionwise_feedforward = PositionwiseFeedforwardLayer(hid_dim,
                                                                     pf_dim,
                                                                     dropout)
        self.dropout = nn.Dropout(dropout)

    def forward(self, src, src_mask):

        #src = [batch size, src len, hid dim]
        #src_mask = [batch size, src len]

        #self attention
        _src, _ = self.self_attention(src, src, src, src_mask)

        #dropout, residual connection and layer norm
        src = self.self_attn_layer_norm(src + self.dropout(_src))

        #src = [batch size, src len, hid dim]

        #positionwise feedforward
        _src = self.positionwise_feedforward(src)

        #dropout, residual and layer norm
        src = self.ff_layer_norm(src + self.dropout(_src))

        #src = [batch size, src len, hid dim]

        return src

class MultiHeadAttentionLayer(nn.Module):
    def __init__(self, hid_dim, n_heads, dropout, device):
        super().__init__()

        assert hid_dim % n_heads == 0

        self.hid_dim = hid_dim
        self.n_heads = n_heads
        self.head_dim = hid_dim // n_heads

        self.fc_q = nn.Linear(hid_dim, hid_dim)
        self.fc_k = nn.Linear(hid_dim, hid_dim)
        self.fc_v = nn.Linear(hid_dim, hid_dim)

        self.fc_o = nn.Linear(hid_dim, hid_dim)

        self.dropout = nn.Dropout(dropout)

        self.scale = torch.sqrt(torch.FloatTensor([self.head_dim])).to(device)

    def forward(self, query, key, value, mask = None):

        batch_size = query.shape[0]

        #query = [batch size, query len, hid dim]
        #key = [batch size, key len, hid dim]
        #value = [batch size, value len, hid dim]

        Q = self.fc_q(query)
        K = self.fc_k(key)
        V = self.fc_v(value)

        #Q = [batch size, query len, hid dim]
        #K = [batch size, key len, hid dim]
        #V = [batch size, value len, hid dim]

        Q = Q.view(batch_size, -1, self.n_heads, self.head_dim).permute(0, 2, 1, 3)
        K = K.view(batch_size, -1, self.n_heads, self.head_dim).permute(0, 2, 1, 3)
        V = V.view(batch_size, -1, self.n_heads, self.head_dim).permute(0, 2, 1, 3)

        #Q = [batch size, n heads, query len, head dim]
        #K = [batch size, n heads, key len, head dim]
        #V = [batch size, n heads, value len, head dim]

        energy = torch.matmul(Q, K.permute(0, 1, 3, 2)) / self.scale

        #energy = [batch size, n heads, query len, key len]

        if mask is not None:
            energy = energy.masked_fill(mask == 0, -1e10)

        attention = torch.softmax(energy, dim = -1)

        #attention = [batch size, n heads, query len, key len]

        x = torch.matmul(self.dropout(attention), V)

        #x = [batch size, n heads, query len, head dim]

        x = x.permute(0, 2, 1, 3).contiguous()

        #x = [batch size, query len, n heads, head dim]

        x = x.view(batch_size, -1, self.hid_dim)

        #x = [batch size, query len, hid dim]

        x = self.fc_o(x)

        #x = [batch size, query len, hid dim]

        return x, attention

class PositionwiseFeedforwardLayer(nn.Module):
    def __init__(self, hid_dim, pf_dim, dropout):
        super().__init__()

        self.fc_1 = nn.Linear(hid_dim, pf_dim)
        self.fc_2 = nn.Linear(pf_dim, hid_dim)

        self.dropout = nn.Dropout(dropout)

    def forward(self, x):

        #x = [batch size, seq len, hid dim]

        x = self.dropout(torch.relu(self.fc_1(x)))

        #x = [batch size, seq len, pf dim]

        x = self.fc_2(x)

        #x = [batch size, seq len, hid dim]

        return x

class Decoder(nn.Module):
    def __init__(self,
                 output_dim,
                 hid_dim,
                 n_layers,
                 n_heads,
                 pf_dim,
                 dropout,
                 device,
                 max_length = 100):
        super().__init__()

        self.device = device

        self.tok_embedding = nn.Embedding(output_dim, hid_dim)
        self.pos_embedding = nn.Embedding(max_length, hid_dim)

        self.layers = nn.ModuleList([DecoderLayer(hid_dim,
                                                  n_heads,
                                                  pf_dim,
                                                  dropout,
                                                  device)
                                     for _ in range(n_layers)])

        self.fc_out = nn.Linear(hid_dim, output_dim)

        self.dropout = nn.Dropout(dropout)

        self.scale = torch.sqrt(torch.FloatTensor([hid_dim])).to(device)

    def forward(self, trg, enc_src, trg_mask, src_mask):

        #trg = [batch size, trg len]
        #enc_src = [batch size, src len, hid dim]
        #trg_mask = [batch size, trg len]
        #src_mask = [batch size, src len]

        batch_size = trg.shape[0]
        trg_len = trg.shape[1]

        pos = torch.arange(0, trg_len).unsqueeze(0).repeat(batch_size, 1).to(self.device)

        #pos = [batch size, trg len]

        trg = self.dropout((self.tok_embedding(trg) * self.scale) + self.pos_embedding(pos))

        #trg = [batch size, trg len, hid dim]

        for layer in self.layers:
            trg, attention = layer(trg, enc_src, trg_mask, src_mask)

        #trg = [batch size, trg len, hid dim]
        #attention = [batch size, n heads, trg len, src len]

        output = self.fc_out(trg)

        #output = [batch size, trg len, output dim]

        return output, attention

class DecoderLayer(nn.Module):
    def __init__(self,
                 hid_dim,
                 n_heads,
                 pf_dim,
                 dropout,
                 device):
        super().__init__()

        self.self_attn_layer_norm = nn.LayerNorm(hid_dim)
        self.enc_attn_layer_norm = nn.LayerNorm(hid_dim)
        self.ff_layer_norm = nn.LayerNorm(hid_dim)
        self.self_attention = MultiHeadAttentionLayer(hid_dim, n_heads, dropout, device)
        self.encoder_attention = MultiHeadAttentionLayer(hid_dim, n_heads, dropout, device)
        self.positionwise_feedforward = PositionwiseFeedforwardLayer(hid_dim,
                                                                     pf_dim,
                                                                     dropout)
        self.dropout = nn.Dropout(dropout)

    def forward(self, trg, enc_src, trg_mask, src_mask):

        #trg = [batch size, trg len, hid dim]
        #enc_src = [batch size, src len, hid dim]
        #trg_mask = [batch size, trg len]
        #src_mask = [batch size, src len]

        #self attention
        _trg, _ = self.self_attention(trg, trg, trg, trg_mask)

        #dropout, residual connection and layer norm
        trg = self.self_attn_layer_norm(trg + self.dropout(_trg))

        #trg = [batch size, trg len, hid dim]

        #encoder attention
        _trg, attention = self.encoder_attention(trg, enc_src, enc_src, src_mask)

        #dropout, residual connection and layer norm
        trg = self.enc_attn_layer_norm(trg + self.dropout(_trg))

        #trg = [batch size, trg len, hid dim]

        #positionwise feedforward
        _trg = self.positionwise_feedforward(trg)

        #dropout, residual and layer norm
        trg = self.ff_layer_norm(trg + self.dropout(_trg))

        #trg = [batch size, trg len, hid dim]
        #attention = [batch size, n heads, trg len, src len]

        return trg, attention

class Seq2Seq(nn.Module):
    def __init__(self,
                 encoder,
                 decoder,
                 src_pad_idx,
                 trg_pad_idx,
                 device):
        super().__init__()

        self.encoder = encoder
        self.decoder = decoder
        self.src_pad_idx = src_pad_idx
        self.trg_pad_idx = trg_pad_idx
        self.device = device

    def make_src_mask(self, src):

        #src = [batch size, src len]

        src_mask = (src != self.src_pad_idx).unsqueeze(1).unsqueeze(2)

        #src_mask = [batch size, 1, 1, src len]

        return src_mask

    def make_trg_mask(self, trg):

        #trg = [batch size, trg len]

        trg_pad_mask = (trg != self.trg_pad_idx).unsqueeze(1).unsqueeze(2)

        #trg_pad_mask = [batch size, 1, 1, trg len]

        trg_len = trg.shape[1]

        trg_sub_mask = torch.tril(torch.ones((trg_len, trg_len), device = self.device)).bool()

        #trg_sub_mask = [trg len, trg len]

        trg_mask = trg_pad_mask & trg_sub_mask

        #trg_mask = [batch size, 1, trg len, trg len]

        return trg_mask

    def forward(self, src, trg):

        #src = [batch size, src len]
        #trg = [batch size, trg len]

        src_mask = self.make_src_mask(src)
        trg_mask = self.make_trg_mask(trg)

        #src_mask = [batch size, 1, 1, src len]
        #trg_mask = [batch size, 1, trg len, trg len]

        enc_src = self.encoder(src, src_mask)

        #enc_src = [batch size, src len, hid dim]

        output, attention = self.decoder(trg, enc_src, trg_mask, src_mask)

        #output = [batch size, trg len, output dim]
        #attention = [batch size, n heads, trg len, src len]

        return output, attention

def translate_sentence(sentence, model, device, max_len = 50):

    model.eval()
    tokens = tokenizer(sentence)

    # if isinstance(sentence, str):
        # nlp = spacy.load('de')
        # tokens = [token.text.lower() for token in nlp(sentence)]
    # else:
        # tokens = [token.lower() for token in sentence]

    tokens = ['<sos>'] + tokens + ['<eos>']

    src_indexes = [src_field_stoi[token] for token in tokens]

    src_tensor = torch.LongTensor(src_indexes).unsqueeze(0).to(device)

    src_mask = model.make_src_mask(src_tensor)

    with torch.no_grad():
        enc_src = model.encoder(src_tensor, src_mask)

    trg_indexes = [trg_field_stoi['<sos>']]

    for i in range(max_len):

        trg_tensor = torch.LongTensor(trg_indexes).unsqueeze(0).to(device)

        trg_mask = model.make_trg_mask(trg_tensor)

        with torch.no_grad():
            output, attention = model.decoder(trg_tensor, enc_src, trg_mask, src_mask)

        pred_token = output.argmax(2)[:,-1].item()

        trg_indexes.append(pred_token)

        if pred_token == trg_field_stoi['<eos>']:
            break

    trg_tokens = [trg_field_itos[i] for i in trg_indexes]

    return trg_tokens[1:], attention

device = "cuda:0" if torch.cuda.is_available() else "cpu"

fn = pathlib.Path(__file__).parent / 'vocab.pt'

src_field_stoi, trg_field_stoi, src_field_itos, trg_field_itos = torch.load(fn)

INPUT_DIM = 50 # len(data.TEXT.vocab)
OUTPUT_DIM = 51 # len(data.LABEL.vocab)
HID_DIM = 256
ENC_LAYERS = 3
DEC_LAYERS = 3
ENC_HEADS = 8
DEC_HEADS = 8
ENC_PF_DIM = 512
DEC_PF_DIM = 512
ENC_DROPOUT = 0.1
DEC_DROPOUT = 0.1

enc = Encoder(INPUT_DIM,
              HID_DIM,
              ENC_LAYERS,
              ENC_HEADS,
              ENC_PF_DIM,
              ENC_DROPOUT,
              device)

dec = Decoder(OUTPUT_DIM,
              HID_DIM,
              DEC_LAYERS,
              DEC_HEADS,
              DEC_PF_DIM,
              DEC_DROPOUT,
              device)

SRC_PAD_IDX = 1
TRG_PAD_IDX = 1

model = Seq2Seq(enc, dec, SRC_PAD_IDX, TRG_PAD_IDX, device).to(device)

fn = pathlib.Path(__file__).parent / 'final-model.pt'
model.load_state_dict(torch.load(fn))



from ._hangle import decompose
from ._hangle import character_is_korean

def levenshtein(s1, s2, cost=None):
    # based on Wikipedia/Levenshtein_distance#Python
    if len(s1) < len(s2):
        return levenshtein(s2, s1)

    if len(s2) == 0:
        return len(s1)
    
    if not cost:
        cost = {}
    
    def get_cost(c1, c2, cost):
        return 0 if (c1 == c2) else cost.get((c1, c2), 1)
    
    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1 # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1       # than s2
            substitutions = previous_row[j] + get_cost(c1, c2, cost)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    
    return previous_row[-1]

def jamo_levenshtein(s1, s2):
    if len(s1) < len(s2):
        return jamo_levenshtein(s2, s1)

    if len(s2) == 0:
        return len(s1)
    
    def get_jamo_cost(c1, c2):
        c1_is_korean, c2_is_korean = character_is_korean(c1), character_is_korean(c2)
        if c1_is_korean and c2_is_korean:
            return 0 if (c1 == c2) else levenshtein(decompose(c1), decompose(c2))/3
        elif not (c1_is_korean) and not (c2_is_korean):
            return 0 if (c1 == c2) else 1
        else:
            return 1

    previous_row = range(len(s2) + 1)
    for i, c1 in enumerate(s1):
        current_row = [i + 1]
        for j, c2 in enumerate(s2):
            insertions = previous_row[j + 1] + 1 # j+1 instead of j since previous_row and current_row are one character longer
            deletions = current_row[j] + 1       # than s2
            substitutions = previous_row[j] + get_jamo_cost(c1, c2)
            current_row.append(min(insertions, deletions, substitutions))
        previous_row = current_row
    
    return previous_row[-1]


from . import CustomModel # seq2seq module
from . import Utils # other functions
from Utils import *

def model_only(sentence):
  text = spacer(sentence)
  text_list = tokenizer(text)
  id_list, wrong_list = check_error(text_list)
  final_wrong, pceq = clean_w_pceq(id_list, text_list)
  corrected = seq2seq(final_wrong)
  output = correct(text_list, id_list, corrected, pceq)
  output = sum(output)
  return output

def both(sentence):
  text = spacer(sentence)
  text_list = tokenizer(text)
  id_list, wrong_list = check_error(text_list)
  final_wrong, pceq = clean_w_pceq(id_list, text_list)
  corrected = seq2seq(final_wrong)
  corrected2 = edit_distance_04(final_wrong)
  result = compare(final_wrong, corrected, corrected2)
  output = correct(text_list, id_list, result, pceq)
  output = sum(output)
  return output

def model_only_word(word):
  result = seq2seq([word])
  suggestion = result[0]
  # print("Suggestion: ", suggestion)
  return suggestion

def edit_only_word(word):
  result = edit_distance_04([word])
  suggestions = ""

  for i in range(len(result[0])):
    if i == len(result[0]) - 1:
      suggestions += result[0][i][0]
    else:
      suggestions += result[0][i][0] + ", "

  # print("Suggestions: ", suggestions)
  return suggestions

def both_word(word):
  model_prediction = model_only_word(word)
  edit_prediction = edit_only_word(word)
  # print("Model suggestion: ", model_prediction)
  # print("Edit Distance suggestion: ", edit_prediction)
  return model_prediction, edit_prediction



import re
import numpy as np

kor_begin, kor_end = (44032, 55203)
jaum_begin, jaum_end = (912593, 12622)
moum_begin, moum_end = (12623, 12643)
chosung_base  = 588
jungsung_base = 28

chosung_list = [ 'ㄱ', 'ㄲ', 'ㄴ', 'ㄷ', 'ㄸ', 'ㄹ', 'ㅁ', 'ㅂ', 'ㅃ', 
        'ㅅ', 'ㅆ', 'ㅇ' , 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

jungsung_list = ['ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ', 
        'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ', 'ㅙ', 'ㅚ', 
        'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ', 
        'ㅡ', 'ㅢ', 'ㅣ']

jongsung_list = [
    ' ', 'ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ', 'ㄶ', 'ㄷ',
        'ㄹ', 'ㄺ', 'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ', 'ㅀ', 
        'ㅁ', 'ㅂ', 'ㅄ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅊ', 
        'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

jaum_list = ['ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ', 'ㄶ', 'ㄷ', 'ㄸ', 'ㄹ', 
              'ㄺ', 'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ', 'ㅀ', 'ㅁ', 'ㅂ', 
              'ㅃ', 'ㅄ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

moum_list = ['ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ', 'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ', 
              'ㅙ', 'ㅚ', 'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ', 'ㅡ', 'ㅢ', 'ㅣ']


def compose(chosung, jungsung, jongsung):
    return chr(kor_begin + chosung_base * chosung_list.index(chosung) + jungsung_base * jungsung_list.index(jungsung) + jongsung_list.index(jongsung))

def decompose(c):    
    if not character_is_korean(c):
        return None
    i = ord(c)
    if (jaum_begin <= i <= jaum_end):
        return (c, ' ', ' ')
    if (moum_begin <= i <= moum_end):
        return (' ', c, ' ')    
    i -= kor_begin
    cho  = i // chosung_base
    jung = ( i - cho * chosung_base ) // jungsung_base 
    jong = ( i - cho * chosung_base - jung * jungsung_base )    
    return (chosung_list[cho], jungsung_list[jung], jongsung_list[jong])

def character_is_korean(c):
    i = ord(c)
    return (kor_begin <= i <= kor_end) or (jaum_begin <= i <= jaum_end) or (moum_begin <= i <= moum_end)


from gensim.corpora.dictionary import Dictionary
from chatspace import ChatSpace

import os
import sys
import re
import pathlib
import pandas as pd

sys.path.append(os.path.dirname(os.path.abspath(os.path.dirname(__file__))))

from ._index import LevenshteinIndex # edit distance
from . import CustomModel
from . import Jamo

# 사전 & 사후 spacing
def spacer(text):
  spacer = ChatSpace()
  result = spacer.space(text)
  return result

# spacing 단위로 tokenize
def tokenizer(text):
  return list(text.split(" "))

# 사전 로드
fn = pathlib.Path(__file__).parent / 'dictionary.txt'
loaded_dct = Dictionary.load_from_text(fn)

# 표기 오류가 있는 단어 indices, words return
def check_error(word_list):
  wrong_ids = []
  wrong_words = []
  pattern = re.compile('[ㄱ-ㅣa-zA-Z0-9]+')

  for i in range(len(word_list)):
    word = word_list[i]
    matched = pattern.search(word)
    hangul = re.compile('[^가-힣]+')
    word = hangul.sub('', word) # 특수문자, 이모지 제거

    # 한글 자모, 영어, 숫자가 포함된 경우는 패스
    if matched != None:
      pass
    elif word == "": # 특수문자만 있었던 경우
      pass
    elif word in loaded_dct.token2id:
      pass
    else:
      wrong_ids.append(i)
      wrong_words.append(word_list[i])

  return wrong_ids, wrong_words

# 특수문자는 다 제거하지만, 끝에 .,!?가 오는 경우에는 모든 처리가 끝난 후 붙여줄 것
# return 특수문자 제거된 표기 오류 리스트, 처리가 끝난 후 붙여줄 mark와 index를 담은 리스트

def clean_w_pceq(id_list, text_list):
  pattern = re.compile('[.,!?]')
  cleaned = []
  pceq = []

  for idx in id_list:
    text = text_list[idx]
    matched = pattern.search(text)

    if matched == None:
      pass
    elif text.endswith('.'):
      pceq.append([idx, "."])
    elif text.endswith(','):
      pceq.append([idx, ","])
    elif text.endswith('!'):
      pceq.append([idx, "!"])
    elif text.endswith('?'):
      pceq.append([idx, "?"])
    else:
      pass

    hangul = re.compile('[^가-힣]+')
    new_text = hangul.sub('', text) # 특수문자, 이모지 제거

    cleaned.append(new_text)

  return cleaned, pceq

def split_text(text):
  t = ""
  for item in text:
    d = Jamo.split_syllables(item)
    if len(d) != 3:
      d = d + "P"
      t = t + d
    else:
      t = t + d

  return t

def join_text(text):
  new = list(filter(lambda a: a != 'p', text))
  new = list(filter(lambda a: a != '<eos>', new))

  join = Jamo.join_jamos(new)

  return join

# Custom Model에서 구하기
def seq2seq(text_list):
  model = CustomModel.model
  prediction_list = []

  for item in text_list:
    text = split_text(item) # 오류 있는 단어 tokenizing and padding
    prediction, attention = CustomModel.translate_sentence(text, model, CustomModel.device)
    prediction_list.append(prediction)

  for i in range(len(prediction_list)):
    joined = join_text(prediction_list[i]) # 다시 tockenizing and padding 풀기
    prediction_list[i] = joined

  return prediction_list

fn = pathlib.Path(__file__).parent / 'clean_dataset.csv'
default_word = pd.read_csv(fn, names = ['num', 'words'])
default_word = list(default_word['words'])[1:]

# 빠른 Edit Distance 도출을 위한 indexer
indexer = LevenshteinIndex(default_word, verbose=True)

# Edit Distance로 구하기
# 최대 distance 0.4로 설정
def edit_distance_04(text_list):
    """
    output example:
    [('그래', 0),
    ('그대', 0.3333333333333333),
    ('그려', 0.3333333333333333)]
    """

    prediction_list = []
    for item in text_list:
        result = indexer.jamo_levenshtein_search(item, max_distance=0.4)
        prediction_list.append(result)

    return prediction_list

# Edit Distance로 구하기
# 최대 distance 1.0로 설정
def edit_distance_10(text_list):
    """
    output example:
    [('그래', 0),
    ('그대', 0.3333333333333333),
    ('그려', 0.3333333333333333)]
    """

    prediction_list = []
    for item in text_list:
        result = indexer.jamo_levenshtein_search(item, max_distance=1)
        prediction_list.append(result)

    return prediction_list

def compare(wrong_list, model_list, edit_list):
    final_output = []

    for i in range(len(model_list)):
        if edit_list[i] == []: # edit distance로 예측한 결과가 없는 경우
            final_output.append(model_list[i])

        else:
            size = len(edit_list[i])

            if wrong_list[i] == model_list[i]: # model이 원래 text 그대로 output을 낸 경우
                edit_only_word = "["
                for j in range(size):
                    if j == (size-1):
                        edit_only_word = edit_only_word + str(edit_list[i][j][0]) + "]"
                    else:
                        edit_only_word = edit_only_word + str(edit_list[i][j][0]) + ", "
                final_output.append(edit_only_word)

            else:
                same_check = False
                for j in range(size):
                    if edit_list[i][j][0] == model_list[i]:
                        same_check = True

                if same_check == True:
                    final_output.append(model_list[i])

                else:
                    if model_list[i] in loaded_dct.token2id:
                        final_output.append(model_list[i])
                    else:
                        edit_only_word = "["
                        for j in range(size):
                            if j == (size-1):
                                edit_only_word = edit_only_word + str(edit_list[i][j][0]) + "]"
                            else:
                                edit_only_word = edit_only_word + str(edit_list[i][j][0]) + ", "
                        final_output.append(edit_only_word)

    return final_output

# 오류 교정된 단어 기존 리스트에 대체해주기 (문장부호 포함)
# All inputs are lists
def correct(origin, idx, corrected, pceq):
  count = 0
  for i in idx:
    origin[i] = corrected[count]
    count += 1

  if len(pceq) == 0:
    pass
  else:
    for case in pceq:
      idx = case[0]
      mark = case[1]
      origin[idx] = origin[idx] + mark

  return origin

# 문장으로 합치기
def sum(corrected):
  result = ""
  length = len(corrected)

  for i in range(length):
    if i != length - 1:
      result += corrected[i] + " "
    else: # 끝에는 No spacing
      result += corrected[i]

  return result

#늘어지는 말들 줄이기
kor_begin = 44032
kor_end = 55203
chosung_base = 588
jungsung_base = 28
jaum_begin = 12593
jaum_end = 12622
moum_begin = 12623
moum_end = 12643

chosung_list = [ 'ㄱ', 'ㄲ', 'ㄴ', 'ㄷ', 'ㄸ', 'ㄹ', 'ㅁ', 'ㅂ', 'ㅃ',
        'ㅅ', 'ㅆ', 'ㅇ' , 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

jungsung_list = ['ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ',
        'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ', 'ㅙ', 'ㅚ',
        'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ',
        'ㅡ', 'ㅢ', 'ㅣ']

jongsung_list = [
    ' ', 'ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ', 'ㄶ', 'ㄷ',
        'ㄹ', 'ㄺ', 'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ', 'ㅀ',
        'ㅁ', 'ㅂ', 'ㅄ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅊ',
        'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

jaum_list = ['ㄱ', 'ㄲ', 'ㄳ', 'ㄴ', 'ㄵ', 'ㄶ', 'ㄷ', 'ㄸ', 'ㄹ',
              'ㄺ', 'ㄻ', 'ㄼ', 'ㄽ', 'ㄾ', 'ㄿ', 'ㅀ', 'ㅁ', 'ㅂ',
              'ㅃ', 'ㅄ', 'ㅅ', 'ㅆ', 'ㅇ', 'ㅈ', 'ㅉ', 'ㅊ', 'ㅋ', 'ㅌ', 'ㅍ', 'ㅎ']

moum_list = ['ㅏ', 'ㅐ', 'ㅑ', 'ㅒ', 'ㅓ', 'ㅔ', 'ㅕ', 'ㅖ', 'ㅗ', 'ㅘ',
              'ㅙ', 'ㅚ', 'ㅛ', 'ㅜ', 'ㅝ', 'ㅞ', 'ㅟ', 'ㅠ', 'ㅡ', 'ㅢ', 'ㅣ']

def compose(chosung, jungsung, jongsung):
    char = chr(
        kor_begin +
        chosung_base * chosung_list.index(chosung) +
        jungsung_base * jungsung_list.index(jungsung) +
        jongsung_list.index(jongsung)
    )
    return char

def decompose(c):
    if not character_is_korean(c):
        return None
    i = ord(c)
    if (jaum_begin <= i <= jaum_end):
        return (c, ' ', ' ')
    if (moum_begin <= i <= moum_end):
        return (' ', c, ' ')

    # decomposition rule
    i -= kor_begin
    cho  = i // chosung_base
    jung = ( i - cho * chosung_base ) // jungsung_base
    jong = ( i - cho * chosung_base - jung * jungsung_base )
    #if jong == ' ': jong = 'e'
    return (chosung_list[cho], jungsung_list[jung], jongsung_list[jong])

def character_is_korean(c):
    i = ord(c)
    return ((kor_begin <= i <= kor_end) or
            (jaum_begin <= i <= jaum_end) or
            (moum_begin <= i <= moum_end))

def make_shorter_word(word):
  de = []
  output = []
  for w in word:
    de.append(decompose(w))

  if de.count(" ")>=2:
    print('case')
    if len(de)==2:
      print(compose(de[0][0], de[1][1], ' '))
    elif len(de)==3:
      print(compose(de[0][0], de[1][1], de[2][0]))
    else:
      print(compose(de[0][0], de[1][1], ' ')+compose(de[2][0], de[3][1], ' '))

  result = ""
  length = len(de)
  for i in range(len(de)): # 모든 글자에 대해
    if de[i][0] != 'ㅇ': # 첫글자 초성이 ㅇ이 아닐때
       if i+2<=length and de[i][1] == de[i+1][1]:
         if i+3<=length and de[i][1] == de[i+2][1]:
           if i+4<=length and de[i][1] == de[i+3][1]:
             output.append([de[i][0], de[i][1], de[i+3][2]])
           else:
             output.append([de[i][0], de[i][1], de[i+2][2]])
         else:
           output.append([de[i][0], de[i][1], de[i+1][2]])
       else:
         output.append(de[i])
    else:
       # 앞글자와 같은 중성이면, pass
      if i!=0 and de[i][1] == de[i-1][1]:
        pass
      elif i+2<=length and de[i][1] == de[i+1][1]:
        if i+3<=length and de[i][1] == de[i+2][1]:
          if i+4<=length and de[i][1] == de[i+3][1]:
            output.append([de[i][0], de[i][1], de[i+3][2]])
          else:
           output.append([de[i][0], de[i][1], de[i+2][2]])
        else:
         output.append([de[i][0], de[i][1], de[i+1][2]])
      else:
        output.append(de[i])
  short = []

  for i in range(len(output)):
    if output[i][0]!=' ' and output[i][1]!=' ':
      result += compose(output[i][0], output[i][1], output[i][2])
    elif output[i][1]==' ':
      short.append(output[i][0])
    elif output[i][0]==' ':
      short.append(output[i][1])

  if len(short)>0:
    if len(short)==2:
      result = compose(short[0], short[1], ' ')
    elif len(short)==3:
      result = compose(short[0], short[1], short[2])
  return result



from setuptools import setup, find_packages

setup(
    name                = 'chatchecker',
    version             = '0.2.4',
    description         = 'chatchecker',
    author              = 'Seoyoung Hong',
    author_email        = 'seoyoungh@khu.ac.kr',
    url                 = 'https://github.com/seoyoungh/ko-chat-checker',
    download_url        = 'https://github.com/seoyoungh/ko-chat-checker',
    install_requires    = ['chatspace'],
    packages            = find_packages(exclude = []),
    keywords            = ['chatchecker'],
    python_requires     = '>=3',
    package_data        = {},
    include_package_data = True,
    zip_safe            = False,
    classifiers         = [
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
    ],
)



from collections import defaultdict
from time import time
from ._string_distance import jamo_levenshtein
from ._string_distance import levenshtein
from ._hangle import decompose
from ._hangle import character_is_korean

class LevenshteinIndex:
    def __init__(self, words=None, levenshtein_distance=levenshtein, verbose=False):
        self._words = {}
        self._index = {} # character to words
        self._cho_index, self._jung_index, self._jong_index = {}, {}, {}
        if words:
            self.indexing(words)
        self.verbose = verbose
        self.jamo_base_distance = 1/3

    def indexing(self, words):
        self._words = words if type(words) == dict else {w:0 for w in words if w}
        self._index = defaultdict(lambda: set())
        self._cho_index = defaultdict(lambda: set())
        self._jung_index = defaultdict(lambda: set())
        self._jong_index = defaultdict(lambda: set())

        for word in words:
            # Indexing for levenshtein
            for c in word:
                self._index[c].add(word)
            # Indexing for jamo_levenshtein
            for c in word:
                if not character_is_korean(c):
                    continue
                cho, jung, jong = decompose(c)
                self._cho_index[cho].add(word)
                self._jung_index[jung].add(word)
                self._jong_index[jong].add(word)

        self._index = dict(self._index)
        self._cho_index = dict(self._cho_index)
        self._jung_index = dict(self._jung_index)
        self._jong_index = dict(self._jong_index)

    def levenshtein_search(self, query, max_distance=1):
        search_time = time()
        similars = defaultdict(int)
        (n, nc) = (len(query), len(set(query)))
        for c in set(query):
            for item in self._index.get(c, {}):
                similars[item] += 1

        if self.verbose:
            print('query={}, candidates={} '.format(query, len(similars)), end='')

        similars = {word for word,f in similars.items()
                    if (abs(n-len(word)) <= max_distance) and (abs(nc - f) <= max_distance)}
        if self.verbose:
            print('-> {}'.format(len(similars)), end='')

        dist = {}
        for word in similars:
            dist[word] = levenshtein(word, query)

        if self.verbose:
            search_time = time() - search_time
            print(', time={:.3} sec.'.format(search_time))

        return sorted(filter(lambda x:x[1] <= max_distance, dist.items()), key=lambda x:x[1])

    def jamo_levenshtein_search(self, query, max_distance=1):
        search_time = time()
        similars = defaultdict(lambda: 0)
        (n, nc) = (len(query), len(set(query)))
        for c in set(query):
            if not character_is_korean(c):
                for item in self._index.get(c, {}):
                    similars[item] += 1
                continue
            cho, jung, jong = decompose(c)
            for item in self._cho_index.get(cho, {}):
                similars[item] += self.jamo_base_distance
            for item in self._jung_index.get(jung, {}):
                similars[item] += self.jamo_base_distance
            for item in self._jong_index.get(jong, {}):
                similars[item] += self.jamo_base_distance

        if self.verbose:
            print('query={}, candidates={} '.format(query, len(similars)), end='')

        similars = {word for word,f in similars.items()
                    if (abs(n-len(word)) <= max_distance) and (abs(nc - f) <= max_distance)}
        if self.verbose:
            print('-> {}'.format(len(similars)), end='')

        dist = {}
        for word in similars:
            dist[word] = jamo_levenshtein(word, query)

        if self.verbose:
            search_time = time() - search_time
            print(', time={:.3} sec.'.format(search_time))

        return sorted(filter(lambda x:x[1] <= max_distance, dist.items()), key=lambda x:x[1])